'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = createAction;
exports.handleActionSuccess = handleActionSuccess;
exports.handleActionBusy = handleActionBusy;
exports.handleActionError = handleActionError;
exports.errorType = errorType;
exports.successType = successType;
exports.busyType = busyType;

var _getActionName = require('./getActionName');

var _getActionName2 = _interopRequireDefault(_getActionName);

var _types = require('./types');

var _constants = require('./constants');

var _storeKeeper = require('./storeKeeper');

var _next = require('./next');

var _listeners = require('./listeners');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function createAction(action) {
  if (typeof action !== 'function') throw new Error('Action must be a function');

  if (action._trackedAction) return action._trackedAction;

  var trackedAction = function trackedAction() {

    var store = (0, _storeKeeper.getStore)();

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var handleSuccess = handleActionSuccess.bind(null, action, args);
    var handleError = handleActionError.bind(null, action, args);
    var handleBusy = handleActionBusy.bind(null, action, args);

    try {
      var actionResponse = action.apply(null, [].concat(args, [store]));
      var isAsync = actionResponse instanceof Promise;

      if (isAsync) {
        handleBusy();

        return actionResponse.then(handleSuccess, function (err) {
          return Promise.reject(handleError(err));
        });
      }

      return handleSuccess(actionResponse);
    } catch (error) {
      handleError(error);
      throw error;
    }
  };

  var actionName = (0, _getActionName2.default)(_constants.NAME_PREFIX, action.name, action);

  trackedAction._successType = actionName;
  trackedAction._busyType = actionName + _constants.BUSY_TYPE_SUFFIX;
  trackedAction._errorType = actionName + _constants.ERROR_TYPE_SUFFIX;

  action._trackedAction = trackedAction;

  return trackedAction;
}

function handleActionSuccess(action, args, payload) {
  var store = (0, _storeKeeper.getStore)();

  store.dispatch({
    trackedAction: action._trackedAction,
    type: action._trackedAction._successType,
    payload: payload
  });

  (0, _listeners.notifySuccessListeners)(action, payload, args);
  (0, _listeners.notifyBusyListeners)(action, false, args);
  (0, _next.resolveWaiters)(action, payload);

  return payload;
}

function handleActionBusy(action, args) {
  var store = (0, _storeKeeper.getStore)();

  store.dispatch({
    trackedAction: action._trackedAction,
    type: action._trackedAction._busyType,
    payload: true
  });

  (0, _listeners.notifyBusyListeners)(action, true, args);
}

function handleActionError(action, args, error) {
  var store = (0, _storeKeeper.getStore)();

  var payload = { error: error, args: args, actionwareError: true };

  store.dispatch({
    trackedAction: action._trackedAction,
    type: action._trackedAction._errorType,
    payload: payload
  });

  (0, _listeners.notifyErrorListeners)(action, error, args);
  (0, _next.rejectWaiters)(action, error);

  return error;
}

function errorType(action) {
  return createAction(action)._errorType;
}

function successType(action) {
  return createAction(action)._successType;
}

function busyType(action) {
  return createAction(action)._busyType;
}